import svgwrite


class RGBMixin(object):

    @staticmethod
    def hex_to_rgb(hex_):
        """Convert hex color to svgwrite rgb string format.
        Args:
            hex_ (str): Color as formatted in hex notation
        Returns:
            Color as vgwrite rgb string format.
        """
        hex_ = hex_.lstrip('#')
        r, g, b = tuple(int(hex_[i:i + 2], 16) for i in (0, 2, 4))
        return svgwrite.utils.rgb(r, g, b)


class NameMixin(object):

    def __repr__(self):
        return self.__class__.__name__



import sys
from xml.dom import minidom

import svgwrite

from element import (
    Arrow, Edge, Fill, Geometry, Label, Node, Path, Point, Style, Viewbox
)
from mixins import NameMixin


class Graph(NameMixin):

    def __init__(self, path, output_path=None):
        self.path = path
        if output_path is None:
            self.svg_path = path.replace('.graphml', '.svg')
        else:
            self.svg_path = output_path
        self.xml = minidom.parse(self.path)
        self.nodes = {}
        self.edges = {}
        self.viewbox = None

    @staticmethod
    def get_attrs(element, tag_name, i=0):
        """Get attributes from tagged elementself. If element not found, or element
        has no attributes, return empty dictionary.
        Args:
            element (xml.element): Element to search within.
            tag_name (str): Wanted child element's tag name.
            i (int): index of which element if more than one. Default is 0.
        Returns:
            If successfull, return populated dictionary.
        """
        try:
            items = element.getElementsByTagName(tag_name)[i].attributes.items()
        except IndexError:
            return {}
        else:
            return {k: v for k, v in items}

    @staticmethod
    def get_node_text(node):
        """Get text from node's 'y:NodeLabel' parameter.
        Args:
            node (xml.element): Node to search within.
        Returns:
            If successfull, return text. Else, return None.
        """
        try:
            return (
                node
                .getElementsByTagName('y:NodeLabel')[0]
                .childNodes[0]
                .data
                .strip()
            )
        except IndexError:
            return

    def add_node(
        self, id_, key=None, text='', shape=None,
        height=10.0, width=10.0, x=0.0, y=0.0,
        fill_color="#ffffff", transparent=False,
        border_color="#000000", border_type='line', border_width=1.0,
        geometry=None, label=None, fill=None, border=None,
        **label_kwargs
    ):
        if geometry is None:
            geometry = Geometry(width, height, x, y)
        if fill is None:
            fill = Fill(fill_color, transparent)
        if border is None:
            border = Style(border_color, border_type, border_width)
        if label is None:
            label = Label(**label_kwargs)
        self.nodes[id_] = Node(id_, key, text, shape, label, geometry, fill, border)

    def add_edge(
        self, id_, key=None, source='', target='',
        edge_color="#000000", edge_type='line', width=1.0, smoothed=False,
        sx=0.0, sy=0.0, tx=0.0, ty=0.0, points=[],
        arrow_source=None, arrow_target='delta',
        line_style=None, path=None, arrow=None
    ):
        if line_style is None:
            line_style = Style(edge_color, edge_type, width, smoothed)
        if path is None:
            path = Path(points, sx, sy, tx, ty)
        if arrow is None:
            arrow = Arrow(arrow_source, arrow_target)
        self.edges[id_] = Edge(
            id_, key, self.nodes.get(source), self.nodes.get(target), line_style,
            path, arrow
        )

    def parse_nodes(self):
        nodes = self.xml.getElementsByTagName('node')
        for node in nodes:
            id_ = node.attributes['id'].value
            text = self.get_node_text(node)
            data = self.get_attrs(node, 'data')
            geometry = self.get_attrs(node, 'y:Geometry')
            fill = self.get_attrs(node, 'y:Fill')
            border = self.get_attrs(node, 'y:BorderStyle')
            shape = 'ellipse'
            label = {
                **self.get_attrs(node, 'y:NodeLabel'),
                **self.get_attrs(node, 'y:SmartNodeLabelModelParameter')
            }
            for var_name in Geometry.__init__.__code__.co_varnames:
                if var_name in label:
                    label['l_{}'.format(var_name)] = label.pop(var_name)
            if 'xml:space' in label.keys():
                label['xml_space'] = label.pop('xml:space')
            self.add_node(
                id_, data['key'], text, shape, geometry['height'],
                geometry['width'], geometry['x'], geometry['y'],
                fill['color'], fill['transparent'], border['color'], border['type'],
                border['width'], **label
            )

    def parse_edges(self):
        edges = self.xml.getElementsByTagName('edge')
        for edge in edges:
            id_ = edge.attributes['id'].value
            data = self.get_attrs(edge, 'data')
            source = edge.attributes['source'].value
            target = edge.attributes['target'].value
            ls = self.get_attrs(edge, 'y:LineStyle')
            bend = self.get_attrs(edge, 'y:BendStyle')
            path = self.get_attrs(edge, 'y:Path')
            arrow = self.get_attrs(edge, 'y:Arrows')
            points = [
                Point(**self.get_attrs(edge, 'y:Point', i=i))
                for i, p in enumerate(edge.getElementsByTagName('y:Point'))
            ]
            self.add_edge(
                id_, data['key'], source, target, ls['color'], ls['type'],
                ls['width'], bend['smoothed'], path['sx'], path['sy'], path['tx'],
                path['ty'], points, arrow['source'], arrow['target']
            )

    def create_viewbox(self):
        min_x = sys.maxsize
        min_y = sys.maxsize
        max_x = -sys.maxsize
        max_y = -sys.maxsize
        for node in self.nodes.values():
            if node.coordinates[0] + node.geometry.width > max_x:
                max_x = node.coordinates[0] + node.geometry.width
            if node.coordinates[1] + node.geometry.height > max_y:
                max_y = node.coordinates[1] + node.geometry.height
            if node.coordinates[0] < min_x:
                min_x = node.coordinates[0]
            if node.coordinates[1] < min_y:
                min_y = node.coordinates[1]
        min_x -= Viewbox.padding
        min_y -= Viewbox.padding
        max_x += Viewbox.padding
        max_y += Viewbox.padding
        width = max_x - min_x
        height = max_y - min_y
        self.viewbox = Viewbox(min_x, min_y, width, height)

    def adjust(self):
        if self.viewbox is None:
            self.create_viewbox()
        for node in self.nodes.values():
            node.geometry.translate(-self.viewbox.x, -self.viewbox.y)
        for edge in self.edges.values():
            for point in edge.path.points:
                point.translate(-self.viewbox.x, -self.viewbox.y)

    def draw_svg(self):
        svg = svgwrite.Drawing(filename=self.svg_path, size=self.viewbox.size)
        for edge in self.edges.values():
            path = svg.path(d=edge.d)
            path.stroke(color=edge.color, width=edge.line_style.width)
            path.fill(color="rgb(255,255,255)", opacity=0.0)
            for pos, draw in zip(('start', 'end'), edge.arrow.draw):
                if draw:
                    arrow = svg.marker(
                        id='arrow_{}_{}'.format(edge.id, pos),
                        refX=edge.arrow.width / 2, refY=edge.arrow.ref_y,
                        size=edge.arrow.size,
                        orient='auto',
                        markerUnits='strokeWidth'
                    )
                    # FIXME: arrow needs to clip top edge.arrow.width / 2 is not the solution
                    tip = svg.path(d=edge.arrow.d, fill=edge.color)
                    arrow.add(tip)
                    svg.defs.add(arrow)
                    path['marker-{}'.format(pos)] = arrow.get_funciri()
            svg.add(path)
        for node in self.nodes.values():
            group = svgwrite.container.Group(id=node.key)
            rect = svg.rect(
                insert=node.coordinates, size=node.size, id=node.id, rx=node.rx,
                ry=node.ry
            )
            rect.fill(color=node.color)
            rect.stroke(color=node.border.color, width=node.border.width)
            rect.dasharray(dasharray=node.border.dasharray)
            group.add(rect)
            if node.text:
                label = svg.text(node.text, insert=node.label_coordinates)
                label['alignment-baseline'] = node.label.alignment
                label['xml:space'] = 'preserve'
                label.fill(color=node.label.color)
                group.add(label)
            svg.add(group)
        svg.save()

    def run(self):
        self.parse_nodes()
        self.parse_edges()
        self.adjust()
        self.draw_svg()



import functools

from mixins import NameMixin, RGBMixin


class Point(NameMixin):

    def __init__(self, x, y):
        self.x = float(x)
        self.y = float(y)

    @property
    def coordinates(self):
        return (self.x, self.y)

    def translate(self, x, y=None):
        if y is None:
            y = x
        self.x += x
        self.y += y

    def scale(self, x, y=None):
        if y is None:
            y = x
        self.x *= x
        self.y *= y


class Geometry(Point, NameMixin):

    def __init__(self, width, height, x, y):
        super(Geometry, self).__init__(x, y)
        self.width = float(width)
        self.height = float(height)

    @property
    def size(self):
        return (self.width, self.height)

    @property
    def center(self):
        return (self.width / 2, self.height / 2)


class Viewbox(Geometry, NameMixin):

    padding = 20

    def __init__(self, min_x=0.0, min_y=0.0, width=0.0, height=0.0):
        super(Viewbox, self).__init__(width, height, min_x, min_y)
        self.sx = 2.0
        self.sy = 2.0


class Fill(NameMixin, RGBMixin):

    def __init__(self, color, color2=None, transparent=False, hasColor=False):
        self.color = color
        self.color2 = color2
        self.transparent = transparent
        self.has_color = hasColor


class Style(NameMixin, RGBMixin):

    def __init__(self, color, type_, width=1.0, smoothed=None):
        self._color = color
        self.type = type_
        self.width = float(width)
        self.bend_smoothing = smoothed

    @property
    def dasharray(self):
        if self.type == 'dashed':
            return [15, 2]
        else:
            return None

    @property
    def color(self):
        return self.hex_to_rgb(self._color)


class Label(Geometry, NameMixin, RGBMixin):

    alignment_dict = {'center': 'middle'}

    def __init__(
        self, alignment="center", autoSizePolicy="content",
        fontFamily="Dialog", fontSize=6.0, fontStyle="plain",
        textColor="#000000", visible="true",
        hasBackgroundColor="false", hasLineColor="false",
        horizontalTextPosition="center", verticalTextPosition="center",
        height=10.0, width=10.0, x=0.0, y=0.0,
        iconTextGap=4.0, modelName="custom",
        **kwargs
    ):
        super(Label, self).__init__(width, height, x, y)
        self.alignment = self.alignment_dict.get(alignment, 'middle')
        self.auto_size_policy = autoSizePolicy
        self.font = fontFamily
        self.font_size = fontSize
        self.font_style = fontStyle
        self.text_color = textColor
        self.visible = visible
        self.has_background_color = hasBackgroundColor
        self.has_line_color = hasLineColor
        self.horizontal_text_position = horizontalTextPosition
        self.vertical_text_position = verticalTextPosition
        self.icon_text_gap = iconTextGap
        self.model_name = modelName
        for k, v in kwargs.items():
            setattr(self, k, v)

    @property
    def color(self):
        return self.hex_to_rgb(self.text_color)


class Path(NameMixin):
    def __init__(self, points, sx=0.0, sy=0.0, tx=0.0, ty=0.0):
        self.points = points
        self.sx = sx
        self.sy = sy
        self.tx = tx
        self.ty = ty


class Arrow(Geometry):

    arrows = ['standard', 'delta', 'white_delta', 'plain']

    def __init__(self, source, target, x=0.5, y=0.5, width=3.0, height=2.0):
        super(Arrow, self).__init__(width, height, x, y)
        self.source = source
        self.target = target

    @property
    def draw(self):
        return (self.draw_source, self.draw_target)

    @property
    def draw_source(self):
        return self.source in self.arrows

    @property
    def draw_target(self):
        return self.target in self.arrows

    @property
    def d(self):
        origin_x = 0.0
        origin_y = 0.0
        return 'M{0},{1} L{0},{2} L{3},{4} Z'.format(
            origin_x, origin_y, self.height, self.width, self.ref_y
        )

    @property
    def ref_y(self):
        return self.height / 2


@functools.total_ordering
class Node(RGBMixin):

    _padding = 2

    def __init__(self, id_, key, text, shape, label, geometry, fill, border):
        self.id = id_
        self.text = text
        self.shape = shape
        self.key = key
        self.label = label
        self.geometry = geometry
        self.fill = fill
        self.border = border
        self.rx = 6.0
        self.ry = 6.0

    def __repr__(self):
        return '<Node {}>'.format(self.id)

    def __str__(self):
        return self.text

    def __eq__(self, node):
        if not isinstance(node, Node):
            return TypeError('Can only compare type Node to type Node.')
        return (
            self.geometry.x == node.geometry.x
            and self.geometry.y == node.geometry.y
        )

    def __lt__(self, node):
        if not isinstance(node, Node):
            return TypeError('Can only compare type Node to type Node.')
        if self.geometry.y == node.geometry.y:
            return self.geometry.y < node.geometry.y
        else:
            return self.geometry.x < node.geometry.x

    @property
    def coordinates(self):
        return self.geometry.coordinates

    @property
    def label_coordinates(self):
        return (
            self.geometry.x
            + self.label.width
            + float(getattr(self.label, 'upX', 0)),
            self.geometry.y
            + self.label.height
            + float(getattr(self.label, 'upY', 0))
        )

    @property
    def color(self):
        return self.hex_to_rgb(self.fill.color)

    @property
    def size(self):
        return self.geometry.size

    def location_relation(self, node):
        if not isinstance(node, Node):
            raise TypeError('Argument must be of type Node.')

        if self.geometry.width:
            x_bounds_min = self.geometry.x + (self.geometry.width / 3)
            x_bounds_max = self.geometry.x + (self.geometry.width * 2 / 3)
            if node.geometry.x < x_bounds_min:  # Node left of
                x = self.geometry.x - self._padding
            elif x_bounds_min < node.geometry.x < x_bounds_max:   # Node centered
                x = self.geometry.x + self.geometry.center[0]
            else:  # Node right of
                x = self.geometry.x + self.geometry.width + self._padding
        else:
            x = self.geometry.x

        if self.geometry.height:
            y_bounds_min = self.geometry.height / 3
            y_bounds_max = self.geometry.height * 2 / 3
            if node.geometry.y < y_bounds_min:  # Node above
                y = self.geometry.y + self.height - self._padding
            elif y_bounds_min < node.geometry.y < y_bounds_max:  # Node centered
                y = self.geometry.y + self.geometry.center[1]
            else:  # Node below
                y = self.geometry.y + self._padding
        else:
            y = self.geometry.y

        return (x, y)


class Edge(object):

    def __init__(self, id_, key, source, target, line_style, path, arrow):
        self.id = id_
        self.key = key
        self.source = source
        self.target = target
        self.line_style = line_style
        self.path = path
        self.arrow = arrow

    def __repr__(self):
        return '<Edge {} {}->{}>'.format(self.id, self.source.id, self.target.id)

    @property
    def start_coordinates(self):
        return self.source.location_relation(self.target)

    @property
    def end_coordinates(self):
        return self.target.location_relation(self.source)

    @property
    def color(self):
        return self.line_style.color

    @property
    def d(self):
        data_str = 'M{},{} '.format(*self.start_coordinates)
        for point in self.path.points:
            data_str += 'L{},{} '.format(*point.coordinates)
        data_str += 'L{},{}'.format(*self.end_coordinates)
        return data_str



from setuptools import setup

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='graphml2svg',
    url='https://github.com/alex-fe/Graphml-to-SVG-converter',
    author='Alex Feldman',
    author_email='alexfeldman93@gmail.com',
    packages=['converter'],
    install_requires=['svgwriter'],
    version='0.0.1',
    license='MIT',
    description='YFile .graphml converter to .svg',
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)
